@@ -37,6 +37,12 @@
   #:use-module (shepherd system)
   #:use-module (shepherd args)
   #:use-module (shepherd comm)
+  #:autoload   (ice-9 binary-ports) (put-u8 get-u8)
+  #:autoload   (fibers operations) (choice-operation
+                                    perform-operation
+                                    wrap-operation)
+  #:autoload   (fibers io-wakeup) (wait-until-port-readable-operation)
+  #:autoload   (fibers timers) (sleep-operation)
   #:export (main))
 
 
@@ -68,34 +74,45 @@ socket file at FILE-NAME upon exit of PROC.  Return the values of PROC."
 (define (maybe-signal-port signals)
   "Return a signal port for SIGNALS, using 'signalfd' on GNU/Linux, or #f if
 that is not supported."
-  (catch 'system-error
-    (lambda ()
-      (let ((port (signalfd -1 signals (logior SFD_CLOEXEC SFD_NONBLOCK))))
-        ;; As per the signalfd(2) man page, block SIGNALS.  The tricky bit is
-        ;; that SIGNALS must be blocked for all the threads; new threads will
-        ;; inherit the signal mask, but we must ensure that neither Guile's
-        ;; signal delivery thread nor its finalization thread are already
-        ;; running, because if they do, they are not blocking SIGNALS.  The
-        ;; signal delivery thread is started on the first call to 'sigaction'
-        ;; so we arrange to not call 'sigaction' beforehand; as for the
-        ;; finalization thread, use 'without-automatic-finalization' to
-        ;; temporarily stop it.
-        (without-automatic-finalization
-         (let ((count (length (all-threads))))
-           (if (= 1 count)
-               (begin
-                 (block-signals signals)
-                 port)
-               (begin
-                 (local-output (l10n "warning: \
+  ;; Honor this environment variable for testing purposes: it lets us emulate
+  ;; the behavior on non-Linux systems.
+  (match (getenv "SHEPHERD_DISABLE_SIGNALFD")
+    ((or #f "0" "no")
+     (catch 'system-error
+       (lambda ()
+         (let ((port (signalfd -1 signals (logior SFD_CLOEXEC SFD_NONBLOCK))))
+           ;; As per the signalfd(2) man page, block SIGNALS.  The tricky bit is
+           ;; that SIGNALS must be blocked for all the threads; new threads will
+           ;; inherit the signal mask, but we must ensure that neither Guile's
+           ;; signal delivery thread nor its finalization thread are already
+           ;; running, because if they do, they are not blocking SIGNALS.  The
+           ;; signal delivery thread is started on the first call to 'sigaction'
+           ;; so we arrange to not call 'sigaction' beforehand; as for the
+           ;; finalization thread, use 'without-automatic-finalization' to
+           ;; temporarily stop it.
+           (without-automatic-finalization
+            (let ((count (length (all-threads))))
+              (if (= 1 count)
+                  (begin
+                    (block-signals signals)
+                    port)
+                  (begin
+                    (local-output (l10n "warning: \
 already ~a threads running, disabling 'signalfd' support")
-                               count)
-                 (close-port port)
-                 #f))))))
-    (lambda args
-      (if (= ENOSYS (system-error-errno args))
-          #f
-          (apply throw args)))))
+                                  count)
+                    (close-port port)
+                    #f))))))
+       (lambda args
+         (if (= ENOSYS (system-error-errno args))
+             (begin
+               (local-output
+                (l10n "System lacks support for 'signalfd'; \
+using fallback mechanism."))
+               #f)
+             (apply throw args)))))
+    (_
+     (local-output (l10n "Support for 'signalfd' is disabled."))
+     #f)))
 
 (define (handle-SIGINT)
   "Handle SIGINT by stopping the Shepherd, which means rebooting if we're PID 1."
@@ -186,6 +203,16 @@ already ~a threads running, disabling 'signalfd' support")
         (lambda (key . args)                      ;for Guile 2.2
           (handle-key-and-args-exception key args))))))
 
+(define (get-u8/timeout port timeout)
+  "Read a byte from @var{port}; return that byte, or the EOF object, or
+@code{'timeout} if @var{timeout} has expired and nothing was read."
+  (perform-operation
+   (choice-operation (wrap-operation (wait-until-port-readable-operation port)
+                                     (lambda ()
+                                       (get-u8 port)))
+                     (wrap-operation (sleep-operation timeout)
+                                     (const 'timeout)))))
+
 (define* (run-daemon #:key (config-file (default-config-file))
                      socket-file pid-file signal-port poll-services?)
   (define (signal-thunk signal-port)
@@ -195,15 +222,35 @@ already ~a threads running, disabling 'signalfd' support")
           (let loop ()
             (handle-signal-port signal-port)
             (loop)))
-        (lambda ()
-          ;; When not using signalfd(2), there's always a time window before
-          ;; 'select' during which a handler async can be queued but not
-          ;; executed.  Work around it by exiting 'select' every few seconds.
-          (let loop ()
-            (sleep (if poll-services? 0.5 30))
-            (when poll-services?
-              (check-for-dead-services))
-            (loop)))))
+        ;; When not using signalfd(2), register a signal handler.  The handler
+        ;; cannot safely send a message to, say, the process monitor, because
+        ;; handlers run as "asyncs", which may be called anytime, including
+        ;; possibly when the message recipient's fiber was active, leading to
+        ;; a deadlock.  To address that, the handler writes to a "self pipe";
+        ;; a fiber reads from the pipe and invokes the actual handler from
+        ;; there.
+        (match (pipe2 (logior O_NONBLOCK O_CLOEXEC))
+          ((input . output)
+           (let ((handler (lambda (signal)
+                            (put-u8 output signal))))
+             (for-each (lambda (signal)
+                         (sigaction signal handler SA_NOCLDSTOP))
+                       %precious-signals)
+             (lambda ()
+               ;; There's a time window before blocking on 'epoll_wait' during
+               ;; which a handler async can be queued but not executed.  Work
+               ;; around it by exiting Fibers' 'epoll_wait' call periodically.
+               ;;
+               ;; TODO: Avoid this race condition by having Fibers use
+               ;; 'epoll_waitp' or 'pselect' and run asyncs before blocking.
+               (let loop ()
+                 (match (get-u8/timeout input (if poll-services? 0.5 30))
+                   ('timeout
+                    (when poll-services?
+                      (check-for-dead-services)))
+                   (signal
+                    ((signal-handler signal))))
+                 (loop))))))))
 
   ;; We might have file descriptors inherited from our parent, as well as file
   ;; descriptors wrongfully opened by Guile or Fibers (see
@@ -440,32 +487,28 @@ fork in the child process."
           (false-if-exception
            (dynamic-link (string-append %pkglibdir "/crash-handler"))))
 
-        ;; Install signal handlers for everything but SIGCHLD, which is taken
-        ;; care of in (shepherd services).
-        (for-each (lambda (signal)
-                    (sigaction signal (signal-handler signal)))
-                  (delete SIGCHLD %precious-signals))
-
         ;; Run Fibers in such a way that it does not create any POSIX thread,
         ;; because POSIX threads and 'fork' cannot be used together.
         (run-fibers
          (lambda ()
-           (with-service-registry
+           (with-process-monitor
+             (with-service-registry
 
-             ;; Register and start the 'root' service.
-             (register-services (list root-service))
-             (start-service root-service)
+               ;; Register and start the 'root' service.
+               (register-services (list root-service))
+               (start-service root-service)
 
-             (with-process-monitor
                ;; Replace the default 'system*' binding with one that
                ;; cooperates instead of blocking on 'waitpid'.  Replace
                ;; 'primitive-load' (in C as of 3.0.9) with one that does
-               ;; not introduce a continuation barrier.
+               ;; not introduce a continuation barrier.  Replace 'sleep' to
+               ;; avoid blocking in user code such as 'start' methods.
                (replace-core-bindings!
                 (system* (lambda command
                            (spawn-command command)))
                 (system spawn-shell-command)
-                (primitive-load primitive-load*))
+                (primitive-load primitive-load*)
+                ((@ (guile) sleep) (@ (fibers) sleep)))
 
                (run-daemon #:socket-file socket-file
                            #:config-file (or config-file (default-config-file))
