@@ -1,6 +1,6 @@
 ;; support.scm -- Various support facilities, used by herd and shepherd.
 ;; Copyright (C) 2014 A.Sassmannshausen <alex.sassmannshausen@gmail.com>
-;; Copyright (C) 2013-2014, 2016, 2018, 2020, 2022, 2023 Ludovic Courtès <ludo@gnu.org>
+;; Copyright (C) 2013-2014, 2016, 2018, 2020, 2022-2024 Ludovic Courtès <ludo@gnu.org>
 ;; Copyright (C) 2002, 2003 Wolfgang Jährling <wolfgang@pro-linux.de>
 ;; Copyright (C) 2016 Mathieu Lirzin <mthl@gnu.org>
 ;; Copyright (C) 2018 Danny Milosavljevic <dannym@scratchpost.org>
@@ -22,6 +22,7 @@
 
 (define-module (shepherd support)
   #:use-module (shepherd config)
+  #:use-module (shepherd system)                 ;for 'O_CLOEXEC' on Guile 2.2
   #:autoload   (shepherd colors) (color-output? color colorize-string)
   #:use-module (ice-9 match)
   #:use-module (ice-9 format)
@@ -89,12 +90,23 @@
      (write args)
      (newline))))
 
-;; Assert that expression EXPR does not evaluate to `#f'.
-(define-syntax-rule (assert EXPR)
-  (and (not EXPR)
-       (begin
-	 (local-output (l10n "Assertion ~a failed.") 'EXPR)
-	 (throw 'assertion-failed))))
+(define-syntax assert
+  (lambda (s)
+    "Assert that expression @var{exp} does not evaluate to @code{#f}."
+    (syntax-case s ()
+      ((_ exp)
+       (let ((location (and=> (syntax-source #'exp)
+                              (lambda (properties)
+                                (let ((file (assq-ref properties 'filename))
+                                      (line (assq-ref properties 'line)))
+                                  (if (and file line)
+                                      (string-append file ":"
+                                                     (number->string line))
+                                      "<unknown-location>"))))))
+         #`(unless exp
+             (local-output (l10n "~a: Assertion ~s failed.")
+                           #,location 'exp)
+             (throw 'assertion-failed)))))))
 
 (define-syntax-rule (let-loop loop ((variable value) ...)
                               body ...)
@@ -214,7 +226,8 @@ output port, and PROC's result is returned."
         (catch-system-error (delete-file template))))))
 
 (define* (mkdir-p dir #:optional mode)  ;copied from Guix
-  "Create directory DIR and all its ancestors."
+  "Create directory DIR and all its ancestors.  Ensure DIR itself has the
+given MODE."
   (define absolute?
     (string-prefix? "/" dir))
 
@@ -243,7 +256,9 @@ output port, and PROC's result is returned."
                        (and st (eq? 'directory (stat:type st)))))
                  (loop tail path)
                  (apply throw args))))))
-      (() #t))))
+      (()
+       (when mode
+         (chmod dir mode))))))
 
 (define-syntax-rule (with-directory-excursion dir body ...) ;copied from Guix
   "Run BODY with DIR as the process's current directory."
@@ -291,7 +306,7 @@ output port, and PROC's result is returned."
 (define* (display-version #:optional (program-name (program-name)))
   (local-output "~a (~a) ~a" program-name package-name Version)
   ;; TRANSLATORS: '~a' is a placeholder for the copyright year.
-  (local-output (l10n "Copyright (C) ~a the Shepherd authors") 2023)
+  (local-output (l10n "Copyright (C) ~a the Shepherd authors") 2024)
   (local-output (l10n "\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
 This is free software: you are free to change and redistribute it.
@@ -437,7 +452,7 @@ create a template configuration file if non exists."
 
 ;; Socket directory for the system's instance (PID 1).
 (define %system-socket-dir
-  (string-append %localstatedir "/run/shepherd"))
+  (string-append %runstatedir "/shepherd"))
 
 ;; The directory where the socket resides.
 (define default-socket-dir
@@ -463,17 +478,28 @@ which has essential bindings pulled in."
     (module-use! m (resolve-interface '(shepherd service)))
     m))
 
+(cond-expand
+ (guile-3 #t)
+ (else
+  ;; This is missing in Guile 2.2.
+  (define (call-with-port port proc)
+    (call-with-values
+        (lambda () (proc port))
+      (lambda vals
+        (close-port port)
+        (apply values vals))))))
+
 (define (primitive-load* file)
   ;; Like 'primitive-load', but in Scheme, so that it does not introduce a
   ;; continuation barrier that would prevent code in FILE from suspending.
-  (call-with-input-file file
+  (call-with-port (open file (logior O_CLOEXEC O_RDONLY))
     (lambda (port)
       (let loop ((result *unspecified*))
-       (match (read port)
-         ((? eof-object?)
-          result)
-         (exp
-          (loop (primitive-eval exp))))))))
+        (match (read port)
+          ((? eof-object?)
+           result)
+          (exp
+           (loop (primitive-eval exp))))))))
 
 (define (load-in-user-module file)
   "Load FILE in a fresh user module that has essential bindings pulled in."
