@@ -43,6 +43,18 @@ function cleanup
     rm -f $service2_pid
 }
 
+function within_seconds
+{
+    seconds=$1
+    while [ $seconds -gt 0 ]
+    do
+	"$SHELL" -c "$2" && break
+	sleep 1
+	seconds="$(expr $seconds - 1)"
+    done
+    [ $seconds -gt 0 ] || "$SHELL" -c "$2"
+}
+
 trap cleanup EXIT
 
 cat > "$conf"<<EOF
@@ -73,6 +85,7 @@ cat > "$conf"<<EOF
 	 #:start (make-forkexec-constructor %command2
 					    #:pid-file "$PWD/$service2_pid")
 	 #:stop  (make-kill-destructor)
+         #:respawn-delay 0
 	 #:respawn? #t)))
 
 (define %command3
@@ -119,8 +132,7 @@ service_pid_value="`cat $service_pid`"
 kill "$service_pid_value"
 while kill -0 "$service_pid_value"; do sleep 0.3; done
 # shepherd should notice that the service has stopped within one second
-sleep 1
-$herd status test | grep stopped
+within_seconds 3 "$herd status test | grep stopped"
 
 
 
@@ -133,18 +145,25 @@ test "`cat $PWD/$service2_started`" = "started"
 rm -f "$service2_pid"
 kill $service2_pid_value
 while kill -0 "$service2_pid_value"; do sleep 0.3; done
-# shepherd should notice that the service has stopped, and restart it, within one second
-sleep 1;
-$herd status test2 | grep running
+# shepherd should notice that the service has stopped, and restart it, within seconds
+within_seconds 5 "$herd status test2 | grep 'Last respawned'"
 test "`cat $PWD/$service2_started`" = "started
 started"
 
 
-
 # test if nofiles was set properly
 test -f "$service_nofiles"
 nofiles_value="`cat $service_nofiles`"
-test 1567 -eq $nofiles_value
+case $(uname) in
+    GNU)
+	# On GNU/Hurd, rlimits are maintained in libc and are not inherited
+	# across 'fork'.
+	echo "rlimit not inherited: $nofiles_value; skipping" >&2
+	;;
+    *)
+	test 1567 -eq $nofiles_value
+	;;
+esac
 
 # Try to trigger potential race conditions, when killing a process, between
 # the fork and execv calls.
