@@ -99,12 +99,19 @@ find_file_descriptor ()
     done
 }
 
-# At this point, the log file is open.
-test -n "$(find_file_descriptor "$PWD/$service_log")"
+if test -d "/proc/$$/fd"
+then
+    # At this point, the log file is open.
+    test -n "$(find_file_descriptor "$PWD/$service_log")"
 
-# Stop the service and ensure the log file has been closed.
-$herd stop test-file-logging
-test -z "$(find_file_descriptor "$PWD/$service_log")"
+    # Stop the service and ensure the log file has been closed.
+    $herd stop test-file-logging
+    test -z "$(find_file_descriptor "$PWD/$service_log")"
+else
+    # On GNU/Hurd, /proc, when mounted, does not expose as much information as
+    # that of Linux; it lacks the 'fd' sub-directory.
+    echo "cannot inspect open file descriptors; skipping" >&2
+fi
 
 rm -f "$service_pid"
 $herd start test-builtin-logging
