@@ -36,11 +36,16 @@ cat > "$conf"<<EOF
 (register-services
  (list (service
 	 '(test)
-	 #:start (const #t)
-	 #:stop  (lambda _
+         ;; Use a real process to exhibit the scenario where
+         ;; the signal handler needs to send a message to
+         ;; the process monitor and/or service registry.
+	 #:start (make-forkexec-constructor '("sleep" "100"))
+	 #:stop  (lambda (pid)
+                   (kill pid SIGKILL)
 		   (call-with-output-file "$stamp"
 		     (lambda (port)
-		       (display "stopped" port))))
+		       (display "stopped" port)))
+                   #f)
 	 #:respawn? #f)))
 (start-service (lookup-service 'test))
 EOF
